/* vi:set cindent tabstop=2 shiftwidth=2: */
/*
 * unistrings - print the strings of printable unicode characters in files
 * Copyright (c) 20001 Martin Hinner <mhi@penguin.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <getopt.h>

void
usage (char *progname);


int
isuni (unsigned short u)
{

  /* US-ASCII */
  if (u >= 0x20 && u <= 0x7E)
    return 1;

  /* ISO-8859-1,2 */
  if (u >= 0x80 && u <= 0xFF)
    return 1;

  /* This is not an unicode character ... */
  return 0;
}

int
main (int argc, char **argv)
{
  int fd;
  unsigned char c;
  int offset;
  unsigned short u1, u2;
  int len1, len2;
  int i;
  unsigned short *uni1, *uni2;
  int minlen;
  int fflg, oflg;
  char *file;
  char ch;

  minlen = 4;
  fflg = 0;
  oflg = 0;

  while ((ch = getopt (argc, argv, "n:of")) != EOF)
    switch ((char) ch)
      {
      case 'f':
	fflg = 1;
	break;
      case 'n':
	minlen = atoi (optarg);
	break;
      case 'o':
	oflg = 1;
	break;
      case '?':
      default:
	usage (argv[0]);
      }

  argc -= optind;
  argv += optind;


  uni1 = (unsigned short *) malloc (minlen * sizeof (unsigned short));
  uni2 = (unsigned short *) malloc (minlen * sizeof (unsigned short));

  if (!uni1 || !uni2)
    {
      fprintf (stderr, "Not enough memory!\n");
      exit (1);
    }

  do
    {

      if (*argv && strcmp (*argv, "-"))
	{
	  file = *argv++;
	  fd = open (file, O_RDONLY);
	}
      else
	{
	  if (*argv)
	    argv++;
	  file = "STDIN";
	  fd = dup (STDIN_FILENO);
	}

      if (fd < 0)
	{
	  perror (file);
		continue;
	}

      offset = 0;
      u1 = u2 = 0xFFFF;
      len1 = len2 = 0;
      while (read (fd, &c, 1) == 1)
	{

	  if (offset % 2 == 0)
	    {
	      u1 &= 0xFF00;
	      u1 |= c;
	      u2 &= 0x00FF;
	      u2 |= c << 8;
	    }
	  else
	    {
	      u1 &= 0x00FF;
	      u1 |= c << 8;
	      u2 &= 0xFF00;
	      u2 |= c;
	    }

	  if (offset >= 2 && offset % 2 == 0)
	    if (isuni (u2))
	      {

		uni2[len2++] = u2;
		if (len2 == minlen)
		  {

		    if (fflg)
		      printf ("%s: ", file);
		    if (oflg)
		      printf ("%07d ", offset + 1 - (2 * minlen));

		    for (i = 0; i < minlen; i++)
		      printf ("%c", uni2[i]);

		    while (read (fd, &u2, 2) == 2)
		      {
			if (isuni (u2))
			  printf ("%c", u2);
			else
			  {
			    offset += 2;
			    len2 = len1 = 0;
			    u1 = 0xFF00 | (u2 >> 8);
			    u2 = 0;
			    break;
			  }
			offset += 2;
		      }
		    printf ("\n");
		  }

	      }
	    else
	      len2 = 0;
	  else if (isuni (u1))
	    {

	      uni1[len1++] = u1;
	      if (len1 == minlen)
		{
		  if (fflg)
		    printf ("%s: ", file);
		  if (oflg)
		    printf ("%07d ", offset + 1 - (2 * minlen));

		  for (i = 0; i < minlen; i++)
		    printf ("%c", uni1[i]);

		  while (read (fd, &u1, 2) == 2)
		    {
		      if (isuni (u1))
			printf ("%c", u1);
		      else
			{
			  offset += 2;
			  len1 = len2 = 0;
			  u2 = 0xFF00 | (u1 >> 8);
			  u1 = 0;
			  break;
			}
		      offset += 2;
		    }
		  printf ("\n");
		}
	    }
	  else
	    len1 = 0;

	  offset++;
	}

      close (fd);

    }
  while (*argv);
  return 0;
}

void
usage (char *progname)
{
  printf ("%s: [-fo] [-n len] [file ...]\n\n\nSend bug reports to <mhi@penguin.cz>\n", progname);
	exit(1);
}
